@@ -1,8 +1,9 @@
 """pytest-bdd public API."""
+from __future__ import annotations
 
 from pytest_bdd.scenario import scenario, scenarios
 from pytest_bdd.steps import given, then, when
 
-__version__ = "5.0.0"
+__version__ = "6.0.1"
 
-__all__ = [given.__name__, when.__name__, then.__name__, scenario.__name__, scenarios.__name__]
+__all__ = ["given", "when", "then", "scenario", "scenarios"]
