@@ -1,17 +1,20 @@
-import io
+from __future__ import annotations
+
 import os.path
 import re
 import textwrap
+import typing
 from collections import OrderedDict
+from typing import cast
 
-from . import types, exceptions
+from . import exceptions, types
 
 SPLIT_LINE_RE = re.compile(r"(?<!\\)\|")
+STEP_PARAM_RE = re.compile(r"<(.+?)>")
 COMMENT_RE = re.compile(r"(^|(?<=\s))#")
 STEP_PREFIXES = [
     ("Feature: ", types.FEATURE),
     ("Scenario Outline: ", types.SCENARIO_OUTLINE),
-    ("Examples: Vertical", types.EXAMPLES_VERTICAL),
     ("Examples:", types.EXAMPLES),
     ("Scenario: ", types.SCENARIO),
     ("Background:", types.BACKGROUND),
@@ -24,8 +27,11 @@ STEP_PREFIXES = [
     ("But ", None),
 ]
 
+if typing.TYPE_CHECKING:
+    from typing import Any, Iterable, Mapping, Match
 
-def split_line(line):
+
+def split_line(line: str) -> list[str]:
     """Split the given Examples line.
 
     :param str|unicode line: Feature file Examples line.
@@ -35,7 +41,7 @@ def split_line(line):
     return [cell.replace("\\|", "|").strip() for cell in SPLIT_LINE_RE.split(line)[1:-1]]
 
 
-def parse_line(line):
+def parse_line(line: str) -> tuple[str, str]:
     """Parse step line to get the step prefix (Scenario, Given, When, Then or And) and the actual step name.
 
     :param line: Line of the Feature file.
@@ -48,7 +54,7 @@ def parse_line(line):
     return "", line
 
 
-def strip_comments(line):
+def strip_comments(line: str) -> str:
     """Remove comments.
 
     :param str line: Line of the Feature file.
@@ -61,7 +67,7 @@ def strip_comments(line):
     return line.strip()
 
 
-def get_step_type(line):
+def get_step_type(line: str) -> str | None:
     """Detect step type by the beginning of the line.
 
     :param str line: Line of the Feature file.
@@ -71,9 +77,10 @@ def get_step_type(line):
     for prefix, _type in STEP_PREFIXES:
         if line.startswith(prefix):
             return _type
+    return None
 
 
-def parse_feature(basedir, filename, encoding="utf-8"):
+def parse_feature(basedir: str, filename: str, encoding: str = "utf-8") -> Feature:
     """Parse the feature file.
 
     :param str basedir: Feature files base directory.
@@ -89,14 +96,13 @@ def parse_feature(basedir, filename, encoding="utf-8"):
         line_number=1,
         name=None,
         tags=set(),
-        examples=Examples(),
         background=None,
         description="",
     )
-    scenario = None
-    mode = None
+    scenario: ScenarioTemplate | None = None
+    mode: str | None = None
     prev_mode = None
-    description = []
+    description: list[str] = []
     step = None
     multiline_step = False
     prev_line = None
@@ -147,48 +153,34 @@ def parse_feature(basedir, filename, encoding="utf-8"):
 
         # Remove Feature, Given, When, Then, And
         keyword, parsed_line = parse_line(clean_line)
+
         if mode in [types.SCENARIO, types.SCENARIO_OUTLINE]:
             tags = get_tags(prev_line)
-            feature.scenarios[parsed_line] = scenario = Scenario(feature, parsed_line, line_number, tags=tags)
+            scenario = ScenarioTemplate(
+                feature=feature,
+                name=parsed_line,
+                line_number=line_number,
+                tags=tags,
+                templated=mode == types.SCENARIO_OUTLINE,
+            )
+            feature.scenarios[parsed_line] = scenario
         elif mode == types.BACKGROUND:
             feature.background = Background(feature=feature, line_number=line_number)
         elif mode == types.EXAMPLES:
             mode = types.EXAMPLES_HEADERS
-            (scenario or feature).examples.line_number = line_number
-        elif mode == types.EXAMPLES_VERTICAL:
-            mode = types.EXAMPLE_LINE_VERTICAL
-            (scenario or feature).examples.line_number = line_number
+            scenario.examples.line_number = line_number
         elif mode == types.EXAMPLES_HEADERS:
-            (scenario or feature).examples.set_param_names([l for l in split_line(parsed_line) if l])
+            scenario.examples.set_param_names([l for l in split_line(parsed_line) if l])
             mode = types.EXAMPLE_LINE
         elif mode == types.EXAMPLE_LINE:
-            (scenario or feature).examples.add_example([l for l in split_line(stripped_line)])
-        elif mode == types.EXAMPLE_LINE_VERTICAL:
-            param_line_parts = [l for l in split_line(stripped_line)]
-            try:
-                (scenario or feature).examples.add_example_row(param_line_parts[0], param_line_parts[1:])
-            except exceptions.ExamplesNotValidError as exc:
-                if scenario:
-                    raise exceptions.FeatureError(
-                        f"Scenario has not valid examples. {exc.args[0]}",
-                        line_number,
-                        clean_line,
-                        filename,
-                    )
-                else:
-                    raise exceptions.FeatureError(
-                        f"Feature has not valid examples. {exc.args[0]}",
-                        line_number,
-                        clean_line,
-                        filename,
-                    )
+            scenario.examples.add_example([l for l in split_line(stripped_line)])
         elif mode and mode not in (types.FEATURE, types.TAG):
             step = Step(name=parsed_line, type=mode, indent=line_indent, line_number=line_number, keyword=keyword)
             if feature.background and not scenario:
-                target = feature.background
+                feature.background.add_step(step)
             else:
-                target = scenario
-            target.add_step(step)
+                scenario = cast(ScenarioTemplate, scenario)
+                scenario.add_step(step)
         prev_line = clean_line
 
     feature.description = "\n".join(description).strip()
@@ -198,45 +190,48 @@ def parse_feature(basedir, filename, encoding="utf-8"):
 class Feature:
     """Feature."""
 
-    def __init__(self, scenarios, filename, rel_filename, name, tags, examples, background, line_number, description):
-        self.scenarios = scenarios
-        self.rel_filename = rel_filename
-        self.filename = filename
-        self.name = name
-        self.tags = tags
-        self.examples = examples
-        self.name = name
-        self.line_number = line_number
-        self.tags = tags
-        self.scenarios = scenarios
-        self.description = description
-        self.background = background
+    def __init__(
+        self,
+        scenarios: OrderedDict,
+        filename: str,
+        rel_filename: str,
+        name: str | None,
+        tags: set,
+        background: Background | None,
+        line_number: int,
+        description: str,
+    ) -> None:
+        self.scenarios: dict[str, ScenarioTemplate] = scenarios
+        self.rel_filename: str = rel_filename
+        self.filename: str = filename
+        self.tags: set = tags
+        self.name: str | None = name
+        self.line_number: int = line_number
+        self.description: str = description
+        self.background: Background | None = background
 
 
-class Scenario:
+class ScenarioTemplate:
+    """A scenario template.
 
-    """Scenario."""
+    Created when parsing the feature file, it will then be combined with the examples to create a Scenario."""
 
-    def __init__(self, feature, name, line_number, example_converters=None, tags=None):
-        """Scenario constructor.
+    def __init__(self, feature: Feature, name: str, line_number: int, templated: bool, tags=None) -> None:
+        """
 
-        :param pytest_bdd.parser.Feature feature: Feature.
         :param str name: Scenario name.
         :param int line_number: Scenario line number.
-        :param dict example_converters: Example table parameter converters.
         :param set tags: Set of tags.
         """
         self.feature = feature
         self.name = name
-        self._steps = []
+        self._steps: list[Step] = []
         self.examples = Examples()
         self.line_number = line_number
-        self.example_converters = example_converters
         self.tags = tags or set()
-        self.failed = False
-        self.test_function = None
+        self.templated = templated
 
-    def add_step(self, step):
+    def add_step(self, step: Step) -> None:
         """Add step to the scenario.
 
         :param pytest_bdd.parser.Step step: Step.
@@ -245,56 +240,56 @@ class Scenario:
         self._steps.append(step)
 
     @property
-    def steps(self):
-        """Get scenario steps including background steps.
+    def steps(self) -> list[Step]:
+        background = self.feature.background
+        return (background.steps if background else []) + self._steps
 
-        :return: List of steps.
-        """
-        result = []
-        if self.feature.background:
-            result.extend(self.feature.background.steps)
-        result.extend(self._steps)
-        return result
-
-    @property
-    def params(self):
-        """Get parameter names.
-
-        :return: Parameter names.
-        :rtype: frozenset
-        """
-        return frozenset(sum((list(step.params) for step in self.steps), []))
-
-    def get_example_params(self):
-        """Get example parameter names."""
-        return set(self.examples.example_params + self.feature.examples.example_params)
-
-    def get_params(self, builtin=False):
-        """Get converted example params."""
-        for examples in [self.feature.examples, self.examples]:
-            yield examples.get_params(self.example_converters, builtin=builtin)
-
-    def validate(self):
-        """Validate the scenario.
-
-        :raises ScenarioValidationError: when scenario is not valid
-        """
-        params = self.params
-        example_params = self.get_example_params()
-        if params and example_params and params != example_params:
-            raise exceptions.ScenarioExamplesNotValidError(
-                """Scenario "{}" in the feature "{}" has not valid examples. """
-                """Set of step parameters {} should match set of example values {}.""".format(
-                    self.name, self.feature.filename, sorted(params), sorted(example_params)
+    def render(self, context: Mapping[str, Any]) -> Scenario:
+        background_steps = self.feature.background.steps if self.feature.background else []
+        if not self.templated:
+            scenario_steps = self._steps
+        else:
+            scenario_steps = [
+                Step(
+                    name=step.render(context),
+                    type=step.type,
+                    indent=step.indent,
+                    line_number=step.line_number,
+                    keyword=step.keyword,
                 )
-            )
+                for step in self._steps
+            ]
+        steps = background_steps + scenario_steps
+        return Scenario(feature=self.feature, name=self.name, line_number=self.line_number, steps=steps, tags=self.tags)
+
+
+class Scenario:
+
+    """Scenario."""
+
+    def __init__(self, feature: Feature, name: str, line_number: int, steps: list[Step] = None, tags=None) -> None:
+        """Scenario constructor.
+
+        :param pytest_bdd.parser.Feature feature: Feature.
+        :param str name: Scenario name.
+        :param int line_number: Scenario line number.
+        :param set tags: Set of tags.
+        """
+        if steps is None:
+            steps = []
+        self.feature = feature
+        self.name = name
+        self.steps = steps
+        self.line_number = line_number
+        self.tags = tags or set()
+        self.failed = False
 
 
 class Step:
 
     """Step."""
 
-    def __init__(self, name, type, indent, line_number, keyword):
+    def __init__(self, name: str, type: str, indent: int, line_number: int, keyword: str) -> None:
         """Step constructor.
 
         :param str name: step name.
@@ -303,19 +298,17 @@ class Step:
         :param int line_number: line number.
         :param str keyword: step keyword.
         """
-        self.name = name
-        self.keyword = keyword
-        self.lines = []
-        self.indent = indent
-        self.type = type
-        self.line_number = line_number
-        self.failed = False
-        self.start = 0
-        self.stop = 0
-        self.scenario = None
-        self.background = None
+        self.name: str = name
+        self.keyword: str = keyword
+        self.lines: list[str] = []
+        self.indent: int = indent
+        self.type: str = type
+        self.line_number: int = line_number
+        self.failed: bool = False
+        self.scenario: ScenarioTemplate | None = None
+        self.background: Background | None = None
 
-    def add_line(self, line):
+    def add_line(self, line: str) -> None:
         """Add line to the multiple step.
 
         :param str line: Line of text - the continuation of the step name.
@@ -323,7 +316,7 @@ class Step:
         self.lines.append(line)
 
     @property
-    def name(self):
+    def name(self) -> str:
         """Get step name."""
         multilines_content = textwrap.dedent("\n".join(self.lines)) if self.lines else ""
 
@@ -339,35 +332,42 @@ class Step:
         return "\n".join(lines).strip()
 
     @name.setter
-    def name(self, value):
+    def name(self, value: str) -> None:
         """Set step name."""
         self._name = value
 
-    def __str__(self):
+    def __str__(self) -> str:
         """Full step name including the type."""
         return f'{self.type.capitalize()} "{self.name}"'
 
     @property
-    def params(self):
+    def params(self) -> tuple[str, ...]:
         """Get step params."""
         return tuple(frozenset(STEP_PARAM_RE.findall(self.name)))
 
+    def render(self, context: Mapping[str, Any]):
+        def replacer(m: Match):
+            varname = m.group(1)
+            return str(context[varname])
+
+        return STEP_PARAM_RE.sub(replacer, self.name)
+
 
 class Background:
 
     """Background."""
 
-    def __init__(self, feature, line_number):
+    def __init__(self, feature: Feature, line_number: int) -> None:
         """Background constructor.
 
         :param pytest_bdd.parser.Feature feature: Feature.
         :param int line_number: Line number.
         """
-        self.feature = feature
-        self.line_number = line_number
-        self.steps = []
+        self.feature: Feature = feature
+        self.line_number: int = line_number
+        self.steps: list[Step] = []
 
-    def add_step(self, step):
+    def add_step(self, step: Step) -> None:
         """Add step to the background."""
         step.background = self
         self.steps.append(step)
@@ -377,75 +377,43 @@ class Examples:
 
     """Example table."""
 
-    def __init__(self):
+    def __init__(self) -> None:
         """Initialize examples instance."""
-        self.example_params = []
-        self.examples = []
-        self.vertical_examples = []
-        self.line_number = None
+        self.example_params: list[str] = []
+        self.examples: list[list[str]] = []
+        self.line_number: int | None = None
         self.name = None
 
-    def set_param_names(self, keys):
+    def set_param_names(self, keys: list[str]) -> None:
         """Set parameter names.
 
         :param names: `list` of `string` parameter names.
         """
         self.example_params = [str(key) for key in keys]
 
-    def add_example(self, values):
+    def add_example(self, values: list[str]) -> None:
         """Add example.
 
         :param values: `list` of `string` parameter values.
         """
         self.examples.append(values)
 
-    def add_example_row(self, param, values):
-        """Add example row.
+    def as_contexts(self) -> Iterable[dict[str, Any]]:
+        if not self.examples:
+            return
 
-        :param param: `str` parameter name
-        :param values: `list` of `string` parameter values
-        """
-        if param in self.example_params:
-            raise exceptions.ExamplesNotValidError(
-                f"""Example rows should contain unique parameters. "{param}" appeared more than once"""
-            )
-        self.example_params.append(param)
-        self.vertical_examples.append(values)
+        header, rows = self.example_params, self.examples
 
-    def get_params(self, converters, builtin=False):
-        """Get scenario pytest parametrization table.
+        for row in rows:
+            assert len(header) == len(row)
+            yield dict(zip(header, row))
 
-        :param converters: `dict` of converter functions to convert parameter values
-        """
-        param_count = len(self.example_params)
-        if self.vertical_examples and not self.examples:
-            for value_index in range(len(self.vertical_examples[0])):
-                example = []
-                for param_index in range(param_count):
-                    example.append(self.vertical_examples[param_index][value_index])
-                self.examples.append(example)
-
-        if self.examples:
-            params = []
-            for example in self.examples:
-                example = list(example)
-                for index, param in enumerate(self.example_params):
-                    raw_value = example[index]
-                    if converters and param in converters:
-                        value = converters[param](raw_value)
-                        if not builtin or value.__class__.__module__ in {"__builtin__", "builtins"}:
-                            example[index] = value
-                params.append(example)
-            return [self.example_params, params]
-        else:
-            return []
-
-    def __bool__(self):
+    def __bool__(self) -> bool:
         """Bool comparison."""
-        return bool(self.vertical_examples or self.examples)
+        return bool(self.examples)
 
 
-def get_tags(line):
+def get_tags(line: str | None) -> set[str]:
     """Get tags out of the given line.
 
     :param str line: Feature file text line.
@@ -455,6 +423,3 @@ def get_tags(line):
     if not line or not line.strip().startswith("@"):
         return set()
     return {tag.lstrip("@") for tag in line.strip().split(" @") if len(tag) > 1}
-
-
-STEP_PARAM_RE = re.compile(r"\<(.+?)\>")
