@@ -34,16 +34,22 @@ def given_beautiful_article(article):
     pass
 
 """
+from __future__ import annotations
+
+import typing
 
 import pytest
-from _pytest.fixtures import FixtureDef
+from _pytest.fixtures import FixtureDef, FixtureRequest
 
 from .parsers import get_parser
 from .types import GIVEN, THEN, WHEN
-from .utils import get_caller_module_locals
+from .utils import get_caller_module_locals, setdefault
+
+if typing.TYPE_CHECKING:
+    from typing import Any, Callable
 
 
-def get_step_fixture_name(name, type_):
+def get_step_fixture_name(name: str, type_: str) -> str:
     """Get step fixture name.
 
     :param name: string
@@ -54,7 +60,11 @@ def get_step_fixture_name(name, type_):
     return f"pytestbdd_{type_}_{name}"
 
 
-def given(name, converters=None, target_fixture=None):
+def given(
+    name: Any,
+    converters: dict[str, Callable] | None = None,
+    target_fixture: str | None = None,
+) -> Callable:
     """Given step decorator.
 
     :param name: Step name or a parser object.
@@ -67,7 +77,7 @@ def given(name, converters=None, target_fixture=None):
     return _step_decorator(GIVEN, name, converters=converters, target_fixture=target_fixture)
 
 
-def when(name, converters=None, target_fixture=None):
+def when(name: Any, converters: dict[str, Callable] | None = None, target_fixture: str | None = None) -> Callable:
     """When step decorator.
 
     :param name: Step name or a parser object.
@@ -80,7 +90,7 @@ def when(name, converters=None, target_fixture=None):
     return _step_decorator(WHEN, name, converters=converters, target_fixture=target_fixture)
 
 
-def then(name, converters=None, target_fixture=None):
+def then(name: Any, converters: dict[str, Callable] | None = None, target_fixture: str | None = None) -> Callable:
     """Then step decorator.
 
     :param name: Step name or a parser object.
@@ -93,7 +103,12 @@ def then(name, converters=None, target_fixture=None):
     return _step_decorator(THEN, name, converters=converters, target_fixture=target_fixture)
 
 
-def _step_decorator(step_type, step_name, converters=None, target_fixture=None):
+def _step_decorator(
+    step_type: str,
+    step_name: Any,
+    converters: dict[str, Callable] | None = None,
+    target_fixture: str | None = None,
+) -> Callable:
     """Step decorator for the type and the name.
 
     :param str step_type: Step type (GIVEN, WHEN or THEN).
@@ -104,14 +119,16 @@ def _step_decorator(step_type, step_name, converters=None, target_fixture=None):
     :return: Decorator function for the step.
     """
 
-    def decorator(func):
+    def decorator(func: Callable) -> Callable:
         step_func = func
         parser_instance = get_parser(step_name)
         parsed_step_name = parser_instance.name
 
+        # TODO: Try to not attach to both step_func and lazy_step_func
+
         step_func.__name__ = str(parsed_step_name)
 
-        def lazy_step_func():
+        def lazy_step_func() -> Callable:
             return step_func
 
         step_func.step_type = step_type
@@ -120,7 +137,9 @@ def _step_decorator(step_type, step_name, converters=None, target_fixture=None):
         # Preserve the docstring
         lazy_step_func.__doc__ = func.__doc__
 
-        step_func.parser = lazy_step_func.parser = parser_instance
+        setdefault(step_func, "_pytest_bdd_parsers", []).append(parser_instance)
+        setdefault(lazy_step_func, "_pytest_bdd_parsers", []).append(parser_instance)
+
         if converters:
             step_func.converters = lazy_step_func.converters = converters
 
@@ -136,7 +155,7 @@ def _step_decorator(step_type, step_name, converters=None, target_fixture=None):
     return decorator
 
 
-def inject_fixture(request, arg, value):
+def inject_fixture(request: FixtureRequest, arg: str, value: Any) -> None:
     """Inject fixture into pytest fixture request.
 
     :param request: pytest fixture request
@@ -157,7 +176,7 @@ def inject_fixture(request, arg, value):
     old_fd = request._fixture_defs.get(arg)
     add_fixturename = arg not in request.fixturenames
 
-    def fin():
+    def fin() -> None:
         request._fixturemanager._arg2fixturedefs[arg].remove(fd)
         request._fixture_defs[arg] = old_fd
 
